#!/usr/bin/env python
import logging
import os

import chevron
import mdtraj
from path import Path

from easytrajh5.binary import get_binary
from easytrajh5.select import select_mask, slice_parmed_with_mask
from easytrajh5.struct import get_parmed_from_parmed_or_pdb
from easytrajh5.struct import load_parmed, slice_parmed, get_mdtraj_topology_from_pmd
from easytrajh5.traj import convert_h5_to_dcd_and_pdb

logger = logging.getLogger(__name__)
logging.basicConfig(level=logging.INFO)

color1 = "green"
color2 = "purple"

colorid1 = 7
colorid2 = 11


def get_parmed_and_pdb(filename, is_solvent=True):
    """
    :param filename: str - either a .parmed or .pdb file
    :return: (parmed.Structure, str) - second str is the filename of a PDB file
    """
    filename = Path(filename)
    ext = filename.ext.lower()

    if ext not in [".pdb", ".parmed"]:
        raise ValueError(
            f"Can't process {filename} of type {ext}, only .pdb and .parmed"
        )

    pmd = get_parmed_from_parmed_or_pdb(filename)

    if ext == ".pdb" and is_solvent:
        return pmd, filename

    if not is_solvent:
        pmd = slice_parmed_with_mask(pmd, "not {solvent}")

    pdb = filename.with_suffix(".parmed.pdb")
    pmd.save(pdb, overwrite=True)

    return pmd, pdb


pymol_template = """
set defer_builds_mode, 3
{{load_script}}
hide
select backbone, name c+o+n+h+oxt and not solvent
deselect

{{#is_stick}}
show sticks
{{/is_stick}}

{{^is_stick}}
show sticks
hide sticks, backbone
hide sticks, hydro
show sticks, name CA
set cartoon_flat_sheets, 0
set cartoon_loop_radius, 0.4
show cartoon
{{/is_stick}}

{{#is_sphere}}
show spheres
{{/is_sphere}}

util.cba(29,"{{name}}")

{{#ranks}}
select rank {{ranks}}
color {{color1}}, sele
deselect
{{/ranks}}

{{#ranks2}}
select rank {{ranks2}}
color {{color2}}, sele
deselect
{{/ranks2}}
"""


def pymol(filename, mask, mask2, sphere, stick, keep_solvent):
    name = Path(filename).name
    if filename.endswith(".h5"):
        dcd, pdb = convert_h5_to_dcd_and_pdb(filename, keep_solvent)
        structure, pdb = get_parmed_and_pdb(pdb)
        load_script = f"load {pdb}, {name}\n"
        load_script += f"load_traj {dcd}, {name}\n"
    else:
        structure, pdb = get_parmed_and_pdb(filename, keep_solvent)
        load_script = f"load {pdb}, {name}\n"

    data = {
        "load_script": load_script,
        "is_stick": stick,
        "is_sphere": sphere,
        "name": name,
        "color1": color1,
        "color2": color2,
    }
    if mask:
        i_atoms = select_mask(structure, mask)
        data["ranks"] = "+".join(map(str, i_atoms))
    if mask2:
        i_atoms = select_mask(structure, mask2)
        data["ranks2"] = "+".join(map(str, i_atoms))

    logger.info("Starting pymol")
    pymol_script_fname = Path(filename).with_suffix(".pml")
    with open(pymol_script_fname, "w") as f:
        f.write(chevron.render(template=pymol_template, data=data))
    os.system(f"{get_binary('pymol')} {pymol_script_fname}")


chimera_template = """
~display

{{load_script}}

repr stick
~display H
setattr m stickScale 2

{{#is_sphere}}
~ribbon
repr cpk
display
{{/is_sphere}}

{{#is_stick}}
repr stick
~ribbon
display
{{/is_stick}}

{{#select_str}}
color {{color1}},a,r {{select_str}}
{{/select_str}}

{{#select_str2}}
color {{color2}},a,r {{select_str2}}
{{/select_str2}}

start Side View

"""


def chimera(filename, parmed, mask, mask2, sphere, stick, keep_solvent):
    if filename.endswith(".h5"):
        dcd, pdb = convert_h5_to_dcd_and_pdb(filename, keep_solvent)

        if not parmed:
            raise ValueError("Error: -p/--parmed must be set")
        parmed = Path(parmed)
        amber_parm7 = parmed.with_suffix(".parm7")
        pmd = load_parmed(parmed)
        if not keep_solvent:
            pmd = slice_parmed(pmd, select_mask(pmd, "not {solvent}"))
        pmd.save(amber_parm7, overwrite=True)

        chimera_load = Path(filename).with_suffix(".chimera.load")
        with open(chimera_load, "w") as f:
            f.write("\n".join(["namdprmtopdcd", amber_parm7, dcd, ""]))
        load_script = f"open movie:{chimera_load}"
    else:
        pmd, pdb = get_parmed_and_pdb(filename, keep_solvent)
        load_script = f"open {pdb}"

    data = {
        "load_script": load_script,
        "is_sphere": sphere,
        "is_stick": stick,
        "color1": color1,
        "color2": color2,
    }
    if mask:
        i_atoms = select_mask(pmd, mask)
        atoms = []
        top = get_mdtraj_topology_from_pmd(pmd)
        for r in top.residues:
            names = [a.name for a in r.atoms if a.index in i_atoms]
            if len(names):
                atoms.append(f":{r.index + 1}@{','.join(names)}")
        data["select_str"] = " ".join(atoms)
    if mask2:
        i_atoms = select_mask(pmd, mask2)
        atoms = []
        top = get_mdtraj_topology_from_pmd(pmd)
        for r in top.residues:
            names = [a.name for a in r.atoms if a.index in i_atoms]
            if len(names):
                atoms.append(f":{r.index + 1}@{','.join(names)}")
        data["select_str2"] = " ".join(atoms)

    logger.info("Starting chimera")
    chimera_cmd_fname = Path(filename).with_suffix(".chimera.cmd")
    with open(chimera_cmd_fname, "w") as f:
        f.write(chevron.render(template=chimera_template, data=data))
    os.system(f"{get_binary('chimera')} {chimera_cmd_fname}")


vmd_template = """
# turn on lights 0 and 1
light 0 on
light 1 on
light 2 off
light 3 off

# position the stage and axes
axes location off
stage location off

# position and turn on menus
menu main on
menu graphics on
menu main move 700 60
menu graphics move 700 340

display projection orthographic
display depthcue on
display cuestart 1.5
display cueend 2.75
display cuemode Linear

mol new "{{top}}" type {{top_type}}
mol delrep 0 0
{{#trj}}
mol addfile "{{trj}}" type {{trj_type}}
{{/trj}}


{{#is_sphere}}
mol selection {all}
mol representation VDW
mol addrep top

{{#ranks}}
mol selection {{ranks}}
mol representation VDW
mol color colorid {{colorid1}}
mol addrep top
{{/ranks}}

{{#ranks2}}
mol selection {{ranks2}}
mol representation VDW
mol color colorid {{colorid2}}
mol addrep top
{{/ranks2}}

{{/is_sphere}}


{{^is_sphere}}

mol selection {backbone}
mol representation NewCartoon 0.40 20.00 2.50 0
mol addrep top
mol selection {(not (protein and backbone or water) or name CA) and not name "[0-9]?H.*"}
mol representation CPK 1.4 2 8 6
mol addrep top

{{#ranks}}
mol selection {{ranks}}
mol representation CPK 1.4 2 8 6
mol color colorid {{colorid1}}
mol addrep top
{{/ranks}}

{{#ranks2}}
mol selection {{ranks2}}
mol representation CPK 1.4 2 8 6
mol color colorid {{colorid2}}
mol addrep top
{{/ranks2}}

{{/is_sphere}}
"""


def vmd(filename, mask, mask2, sphere, keep_solvent):
    if filename.endswith(".h5"):
        dcd, pdb = convert_h5_to_dcd_and_pdb(filename, keep_solvent)
        structure, pdb = get_parmed_and_pdb(pdb, keep_solvent)
    else:
        dcd = ""
        structure, pdb = get_parmed_and_pdb(filename, keep_solvent)

    data = {
        "top_type": "pdb",
        "trj_type": "dcd",
        "top": pdb,
        "trj": dcd,
        "is_sphere": sphere,
        "colorid1": colorid1,
        "colorid2": colorid2,
    }
    if mask:
        i_atoms = select_mask(structure, mask)
        data["ranks"] = "{index " + " ".join(map(str, i_atoms)) + "}"
    if mask2:
        i_atoms = select_mask(structure, mask2)
        data["ranks2"] = "{index " + " ".join(map(str, i_atoms)) + "}"

    logger.info("Starting vmd")
    vmd_script_fname = Path(filename).with_suffix(".vmd")
    with open(vmd_script_fname, "w") as f:
        f.write(chevron.render(template=vmd_template, data=data))
    os.system(f"{get_binary('vmd')} -e {vmd_script_fname}")



import json
import logging
import os
import subprocess
import sys
import time
from glob import glob

import numpy as np
from addict import Dict
from deepdiff import DeepDiff
from parmed import unit
from path import Path
from pydash import py_
from rich.pretty import pretty_repr
from ruyaml import YAML

from .quantity import get_dict_from_quantity, get_quantity_from_dict

logger = logging.getLogger(__name__)

yaml = YAML(typ="safe")
yaml.default_flow_style = False


class Encoder(json.JSONEncoder):
    def default(self, o):
        if isinstance(o, np.ndarray):
            return o.tolist()
        if isinstance(o, unit.Quantity):
            return get_dict_from_quantity(o)
        return str(o)


def decode(o):
    if isinstance(o, dict):
        if py_.get(o, "type") == "quantity":
            return get_quantity_from_dict(o)
        else:
            return {k: decode(v) for k, v in o.items()}
    elif isinstance(o, list):
        return [decode(item) for item in o]
    return o


def load_json(f, is_addict=False):
    with open(f) as handle:
        result = decode(json.load(handle))
    if is_addict:
        result = Dict(result)
    return result


def dump_json(o, f):
    check_file_dir(f)
    with open(f, "w") as handle:
        json.dump(o, handle, cls=Encoder)


def load_yaml(f, is_addict=False):
    with open(f) as handle:
        result = decode(yaml.load(handle))
    if is_addict:
        result = Dict(result)
    return result


def load_yaml_dict(f):
    return load_yaml(f, is_addict=True)


def dump_yaml(o, f, mode="w"):
    check_file_dir(f)
    o = json.loads(json.dumps(o, cls=Encoder))
    with open(f, mode) as handle:
        yaml.dump(o, handle)


def get_yaml_str(o, indent=0):
    o = json.loads(json.dumps(o, cls=Encoder, default=str))
    import textwrap
    from io import StringIO

    handle = StringIO()
    yaml.dump(o, handle)
    s = handle.getvalue()
    if indent:
        s = textwrap.indent(s, " " * indent)
    if s.endswith("\n"):
        s = s[:-1]
    return s


def print_yaml(o, indent=2):
    print(get_yaml_str(o, indent=indent))


def ensure_dir(d):
    Path(d).makedirs_p()


def check_file_dir(f):
    ensure_dir(Path(f).abspath().parent)


def clear_dir(d):
    path = Path(d)
    if not path.exists():
        ensure_dir(path)
    elif path.isdir():
        for f in path.files():
            f.remove()
        for d in path.dirs():
            d.rmtree()
    else:
        raise ValueError(f"Error: {d} is not a directory")


def copy_to_dir(src, dest_dir):
    src, dest_dir = Path(src), Path(dest_dir)
    dest_dir.makedirs_p()
    dst = dest_dir / src.name
    if dst.abspath() != src.abspath():
        src.copyfile(dst)


def copy_file(source, target, follow_symlinks=True) -> None:
    source, target = Path(source), Path(target)
    if target.exists():
        return
    check_file_dir(target)
    source.copyfile(target, follow_symlinks=follow_symlinks)


def run(cmd, env=None):
    print(">>>>", cmd)
    subprocess.run(cmd, shell=True, check=True, env=env)


def run_with_output(cmd):
    bytes_output = subprocess.check_output(cmd.split())
    return bytes_output.decode("utf-8", errors="ignore")


def move_up_sub_dir(top_dir, glob_str):
    for f in Path(top_dir).glob(glob_str):
        this_dir = f.parent
        if not this_dir.exists():
            continue
        for sibling_f in this_dir.glob("*"):
            if sibling_f.isfile():
                sibling_f.copy(this_dir.parent)
            if sibling_f.isdir():
                sibling_f.copytree(this_dir.parent, dirs_exist_ok=True)
        this_dir.rmtree_p()


def append_to_key_list(a_dict, key, val):
    if key not in a_dict:
        a_dict[key] = []
    a_dict[key].append(val)


def transfer_glob_str(from_dir, glob_str, to_dir):
    """
    Make copies of files in from_dir/glob_str in the directory to_dir

        >>> transfer_glob_str('source', '**/*.html', 'target')
    """
    from_dir = Path(from_dir)
    to_dir = Path(to_dir)
    for f in glob(from_dir / glob_str, recursive=True):
        rel_dir = Path(f).parent.relpath(from_dir)
        to_rel_dir = to_dir / rel_dir
        copy_to_dir(f, to_rel_dir)


def get_checked_path(f):
    result = Path(f)
    if not result.exists():
        logger.error(f"Error: couldn't find {f}")
        sys.exit(1)
    return result


saved_dirs = []


def push_dir(new_dir):
    new_dir = Path(new_dir)
    new_dir.makedirs_p()
    save_dir = os.getcwd()
    saved_dirs.append(save_dir)
    new_dir.chdir()


def pop_dir():
    if len(saved_dirs):
        save_dir = saved_dirs.pop()
        os.chdir(save_dir)


def get_active_branch_name():
    head_dir = Path(__file__).abspath().parent.parent.parent / ".git" / "HEAD"
    if head_dir.exists():
        with head_dir.open("r") as f:
            content = f.read().splitlines()

        for line in content:
            if line[0:4] == "ref:":
                return line.partition("refs/heads/")[2]
    return ""


def overwrite_yaml(options_yaml, options):
    backup_options_yaml = Path(options_yaml).with_suffix(".bak.yaml")
    while backup_options_yaml.exists():
        backup_options_yaml = backup_options_yaml.with_suffix(".bak.yaml")
    logger.info(f"Updated {options_yaml} -> {backup_options_yaml}")
    os.rename(options_yaml, backup_options_yaml)
    dump_yaml(options, options_yaml)


def is_equal_dict(options1, options2):
    diff = DeepDiff(options2, options1)
    if diff:
        print(diff.pretty())
    return not diff


def load_yaml_from_yaml_url(v):
    tokens = v.replace("yaml://", "").split("/")
    yaml_fname, dict_path = tokens[0:2]
    d = load_yaml_dict(yaml_fname)
    return py_.get(d, dict_path)


def walk_object(o):
    if isinstance(o, dict):
        a_dict = o
        for k, v in a_dict.items():
            if isinstance(v, dict) or isinstance(v, list):
                walk_object(v)
            elif isinstance(v, str):
                if v.startswith("yaml://"):
                    a_dict[k] = load_yaml_from_yaml_url(v)
    elif isinstance(o, list):
        a_list = o
        for i in range(len(a_list)):
            v = a_list[i]
            if isinstance(v, dict) or isinstance(v, list):
                walk_object(v)
            elif isinstance(v, str):
                if v.startswith("yaml://"):
                    a_list[i] = load_yaml_from_yaml_url(v)


def parse_dict_with_yaml_url(o):
    parsed_o = Dict(o)
    walk_object(parsed_o)
    return parsed_o


def repr_lines(o, prefix=""):
    lines = pretty_repr(o).split("\n")
    lines[0] = prefix + lines[0]
    return lines


time_store = []


def tic(msg="No message"):
    ti = time.perf_counter_ns()  # final time
    time_store.append((msg, ti))
    return f"{msg}:started..."


def toc():
    tf = time.perf_counter_ns()  # final time
    if not len(time_store):
        delta_ms = 0
        msg = "no tic"
    else:
        msg, ti = time_store.pop()
        delta_ns = tf - ti
        delta_ms = round(delta_ns / 1000000)
    time_s = get_time_str(delta_ms / 1000)
    return f"{msg}:finished in {time_s}"


def get_time_str(seconds):
    days = seconds // 86400
    seconds -= days * 86400
    hours = seconds // 3600
    seconds -= hours * 3600
    minutes = seconds // 60
    seconds -= minutes * 60
    if days > 0:
        value = "%d:%d:%02d:%02d" % (
            days,
            hours,
            minutes,
            seconds,
        )
    elif hours > 0:
        value = "%d:%02d:%02d" % (
            hours,
            minutes,
            seconds,
        )
    elif minutes > 0:
        value = "%02d:%02d" % (minutes, seconds)
    elif seconds >= 1:
        value = "%.3fs" % seconds
    elif seconds > 0.001:
        value = "%.0fms" % (seconds * 1000)
    else:
        value = "<1ms"
    return value



import logging

import mdtraj

from .traj import EasyTrajH5File

logger = logging.getLogger(__name__)


class TrajectoryManager:
    """
    Abstraction to manage multiple trajectories that
    have the same topologies. The exact frame is then referenced
    by [i_frame, i_traj, align_atom_mask].
    """

    def __init__(
            self,
            paths: [str],
            mode: str = "a",
            atom_mask: str = "not {solvent}",
            is_dry_cache: bool = True,
    ) -> None:
        self.paths = paths
        self.mode = mode
        self.atom_mask = atom_mask
        self.is_dry_cache = is_dry_cache

        self.traj_file_by_i = {}
        self.last_i_frame_traj = None
        self.last_frame = None

    def get_n_trajectories(self) -> int:
        return len(self.paths)

    @property
    def backend(self):
        return EasyTrajH5File

    def get_traj_file(self, i_traj) -> EasyTrajH5File:
        if i_traj not in self.traj_file_by_i:
            path = self.paths[i_traj]
            traj_file = self.backend(
                path,
                self.mode,
                atom_mask=self.atom_mask,
                is_dry_cache=self.is_dry_cache,
            )
            if not len(traj_file.get_dataset_keys()):
                raise FileNotFoundError(f"h5: {path} is empty")
            self.traj_file_by_i[i_traj] = traj_file
        return self.traj_file_by_i[i_traj]

    def get_n_frame(self, i_traj) -> int:
        return self.get_traj_file(i_traj).get_n_frame()

    def read_as_frame_traj(self, i_frame_traj) -> mdtraj.Trajectory:
        if self.last_i_frame_traj == i_frame_traj:
            logger.info(f"same as last frame {i_frame_traj}")
            return self.last_frame

        i_frame, i_traj = i_frame_traj[:2]
        if i_frame < 0:
            i_frame = self.get_n_frame(i_traj) + i_frame

        frame = self.get_traj_file(i_traj).read_frame_as_traj(i_frame)

        self.last_i_frame_traj = i_frame_traj
        self.last_frame = frame

        return frame

    def close(self):
        for h5 in self.traj_file_by_i.values():
            h5.close()



import json
import re
from typing import Union, KeysView, Any

import h5py
import numpy
import numpy as np
from addict import Dict
from path import Path
from rich import box
from rich.console import Console
from rich.pretty import pprint
from rich.table import Table

from .select import parse_number_list


def convert_str_to_bytes(s):
    if not isinstance(s, bytes):
        s = s.encode("ascii")
    return s


def convert_bytes_to_str(b: Union[numpy.bytes_, bytes, str]):
    if isinstance(b, numpy.bytes_):
        b = b.tobytes()
    if not isinstance(b, str):
        b = b.decode()
    return b


class EasyH5File:
    """
    EasyH5File is a convenience class to make h5py.File objects
    easier to use for the following use-cases:

        - easily create and add to fixed/extendable numpy arrays
        - add strings as byte column dataset
        - add JSON-compatible object literals
        - add binary blobs; which allows insertion of binary files
        - handle attributes/dataset access using method-based lookups
        - add diagnostics such as JSON schema and print function

    More importantly, this object allows the internal handle type
    of h5py.File to be easily overrideable with h5pyd.File for
    processing with an HSDS data server.

    The h5py.File is stored as self.handle.

    Note: remember to .flush() or .close() object to force write
    """

    def __init__(self, fname, mode="a"):
        self.fname = fname
        self.mode = mode
        self.handle = self.open(self.fname, mode=self.mode)
        self.is_open = True

    def open(self, fname, mode):
        # We use this method to open the h5py object
        # to allow this method to be overriable
        # with a reference to h5pyd for remote
        # access with an HSDS server
        return h5py.File(fname, mode)

    def flush(self):
        self.handle.flush()

    def close(self):
        if self.is_open:
            self.handle.close()
            self.is_open = False

    def __enter__(self):
        return self

    def __exit__(self, *exc_info):
        self.close()

    def __del__(self):
        self.close()

    def has_dataset(self, key) -> bool:
        return key in self.handle

    def get_dataset_keys(self) -> [str]:
        def get_node_keys(node_key=None):
            result = []
            if node_key is None:
                node = self.handle
            else:
                node = self.handle[node_key]
            for key in node.keys():
                if node_key is not None:
                    leaf_key = f"{node_key}/{key}"
                else:
                    leaf_key = key
                if isinstance(self.handle[leaf_key], h5py.Group):
                    result.extend(get_node_keys(leaf_key))
                else:
                    result.append(leaf_key)
            return result

        return get_node_keys()

    def set_array_dataset(self, key, array_data):
        self.clear_dataset(key)
        self.handle.create_dataset(key, data=array_data)

    def clear_dataset(self, key):
        if key in self.handle:
            del self.handle[key]

    def get_dataset(self, key) -> h5py.Dataset:
        return self.handle[key]

    def delete_dataset(self, key):
        if self.has_dataset(key):
            del self.handle[key]

    def create_extendable_dataset(self, key, frame_shape=(), dtype=numpy.float32):
        self.handle.create_dataset(
            key,
            chunks=True,
            compression="gzip",
            dtype=dtype,
            shape=(0, *frame_shape),
            maxshape=(None, *frame_shape),
        )

    def extend_dataset(self, key, new_frames):
        shape = self.handle[key].shape
        n_frame = shape[0]
        frame_shape = shape[1:]
        new_frames = numpy.array(new_frames)
        if new_frames.shape == frame_shape:
            self.handle[key].resize(n_frame + 1, axis=0)
            self.handle[key][-1] = new_frames
        elif new_frames.shape[1:] == frame_shape:
            self.handle[key].resize(n_frame + len(new_frames), axis=0)
            self.handle[key][n_frame:] = new_frames
        else:
            raise ValueError(
                f"Shape of new frames {new_frames.shape} doesn't match {frame_shape}"
            )

    def set_bytes_dataset(self, key, blob):
        self.clear_dataset(key)
        self.handle.create_dataset(key, data=numpy.array([blob]))

    def get_bytes_dataset(self, key) -> bytes:
        return self.get_dataset(key)[0].tobytes()

    def insert_file_to_dataset(self, key, fname):
        with open(fname, "rb") as f:
            blob = f.read()
        self.set_bytes_dataset(key, blob)

    def extract_file_from_dataset(self, key, fname):
        blob = self.get_bytes_dataset(key)
        with open(fname, "wb") as f:
            f.write(blob)

    def set_str_dataset(self, key, s):
        self.set_bytes_dataset(key, convert_str_to_bytes(s))

    def get_str_dataset(self, key) -> str:
        return convert_bytes_to_str(self.get_bytes_dataset(key))

    def set_json_dataset(self, key, obj):
        self.set_str_dataset(key, json.dumps(obj, default=str))

    def get_json_dataset(self, key) -> Any:
        return json.loads(self.get_str_dataset(key))

    def get_attrs(self, dataset_key=None) -> h5py.AttributeManager:
        """Returns h5py attrs object"""
        return self.get_dataset(dataset_key).attrs if dataset_key else self.handle.attrs

    def get_attr_keys(self, dataset_key=None) -> KeysView:
        return self.get_attrs(dataset_key).keys()

    def has_attr(self, key, dataset_key=None) -> bool:
        return key in self.get_attrs(dataset_key)

    def get_attr(self, key, dataset_key=None) -> Any:
        attrs = self.get_attrs(dataset_key)
        v = attrs[key]
        if isinstance(v, numpy.bytes_):
            v = convert_bytes_to_str(v)
        if isinstance(v, h5py._hl.base.Empty):
            v = None
        if isinstance(v, numpy.integer):
            v = int(v)
        if hasattr(numpy, "float") and isinstance(v, numpy.float):
            v = float(v)
        return v

    def get_attr_dict(self, dataset_key=None) -> dict:
        """Returns h5py attrs as a dict"""
        return {
            k: self.get_attr(key, dataset_key)
            for key in self.get_attr_keys(dataset_key)
        }

    def set_attr(self, key, value, dataset_key=None):
        attrs = self.get_attrs(dataset_key)
        if key in attrs:
            attrs.modify(key, value)
        else:
            attrs.create(key, value)

    def get_dataset_schema(self, key) -> Dict:
        dataset = self.get_dataset(key)
        config = Dict(key=key)
        config.shape = list(dataset.shape)
        if dataset.chunks is not None:
            config.chunks = list(dataset.chunks)
        if dataset.maxshape[0] is None:
            config.is_extensible = True
            config.frame_shape = list(dataset.maxshape[1:])
            config.n_frame = config.shape[0]
        config.dtype = str(dataset.dtype)
        if re.search(r"\|S\d+", config.dtype):
            config.dtype = f"string({config.dtype[2:]})"
        for attr_key in self.get_attr_keys(key):
            config.attr[attr_key] = self.get_attr(attr_key, key)
        return config

    def get_schema(self) -> Dict:
        structure = Dict(datasets=[])
        for key in self.get_dataset_keys():
            structure.datasets.append(self.get_dataset_schema(key))
        for key in self.get_attr_keys():
            structure.attr[key] = self.get_attr(key)
        return structure.to_dict()

    def print_schema(self):
        pprint(self.get_schema())

    def print_dataset_table(self, title=None):
        table = Table(title=title, box=box.SIMPLE)
        table.add_column("dataset")
        table.add_column("shape")
        table.add_column("dtype")
        table.add_column("size (MB)", justify="right")

        def to_mb(n):
            mb = n / 1024**2
            if mb < 0.01:
                return "<1 KB"
            return f"{mb:.2f} MB"

        total = 0
        for key in self.get_dataset_keys():
            dataset = self.get_dataset(key)
            n_byte = dataset.nbytes
            total += n_byte
            table.add_row(key, str(dataset.shape), str(dataset.dtype), to_mb(n_byte))

        table.add_row()
        table.add_row("total", "", "", to_mb(total))

        print()
        console = Console()
        console.print(table)

    def print_dataset(self, key, frames=None, is_json=True):
        dataset = self.get_dataset(key)
        dtype = str(dataset.dtype)

        print(f"     dataset={key}")
        print(f"     dtype={dtype}")

        is_string = re.search(r"\|S\d+", dtype)
        if is_string:
            str_len = int(dtype[2:])
            print()
            if is_json:
                print("---- JSON --------------------")
                pprint(self.get_json_dataset(key))
            else:
                b = self.get_bytes_dataset(key)
                try:
                    s = convert_bytes_to_str(b)
                    print("---- STRING --------------------")
                    if str_len > 2000:
                        print(s[:1000])
                        print("\n....\n")
                        print(s[-1000:])
                    else:
                        print(s)
                except Exception:
                    import sys

                    print("---- BYTES --------------------")
                    if str_len > 2000:
                        sys.stdout.buffer.write(b[:1000])
                        print("\n....")
                        sys.stdout.buffer.write(b[-1000:])
                    else:
                        sys.stdout.buffer.write(b)
        else:
            print(f"     shape={dataset.shape}")
            print()
            if frames is not None:
                print(f"frames({frames})=")
                i_frames = parse_number_list(frames)
                chunk = dataset[i_frames]
            else:
                chunk = dataset[:]
            print(chunk)


def dump_attr_to_h5(h5_fname, value, key):
    path = Path(h5_fname)
    mode = "a" if path.isfile() else "w"  # if the h5 exist
    with h5py.File(path, mode) as f:
        f.attrs[key] = value


def create_dataset_in_h5_file_with_value(h5_file, value, key):
    if isinstance(value, list):
        shape = [len(value)]
    elif isinstance(value, np.ndarray):
        shape = value.shape
    else:
        shape = []
    h5_file.create_dataset(key, maxshape=(None, *shape), data=np.array([value]))


def dump_value_to_h5(h5_fname, value, key):
    """
    Convenience function to quickly append values to a dataset to an h5 file.
    The file will be created if it doesn't exist. Similarly, with the dataset.
    The file will be opened and then closed so that there are no dangling
    handles.
    """
    path = Path(h5_fname)

    if path.isfile():  # if the h5 exist
        with h5py.File(path, "a") as f:
            if key in f.keys():
                old_size = f[key].shape[0]
                f[key].resize(old_size + 1, axis=0)
                f[key][old_size:] = value
            else:
                create_dataset_in_h5_file_with_value(f, value, key)
    else:
        with h5py.File(path, "w") as f:
            create_dataset_in_h5_file_with_value(f, value, key)
